@@ -1,49 +1,28 @@
+from __future__ import annotations
+
+import ast
 import re
-
-try:
-    import ast
-except ImportError:
-    from chameleon import ast25 as ast
-
-try:
-    str = unicode
-except NameError:
-    long = int
-
 from functools import partial
-from copy import copy
 
-from ..program import ElementProgram
-
-from ..namespaces import XML_NS
-from ..namespaces import XMLNS_NS
-from ..namespaces import I18N_NS as I18N
-from ..namespaces import TAL_NS as TAL
-from ..namespaces import METAL_NS as METAL
-from ..namespaces import META_NS as META
-
-from ..astutil import Static
-from ..astutil import Symbol
-from ..astutil import parse
-from ..astutil import marker
-
-from .. import tal
-from .. import tales
-from .. import metal
-from .. import i18n
-from .. import nodes
-
-from ..exc import LanguageError
-from ..exc import ParseError
-from ..exc import CompilationError
-
-from ..utils import decode_htmlentities
-from ..utils import ImportableMarker
-
-try:
-    str = unicode
-except NameError:
-    long = int
+from chameleon import i18n
+from chameleon import metal
+from chameleon import nodes
+from chameleon import tal
+from chameleon.astutil import Static
+from chameleon.astutil import Symbol
+from chameleon.astutil import parse
+from chameleon.exc import CompilationError
+from chameleon.exc import LanguageError
+from chameleon.exc import ParseError
+from chameleon.namespaces import I18N_NS as I18N
+from chameleon.namespaces import META_NS as META
+from chameleon.namespaces import METAL_NS as METAL
+from chameleon.namespaces import TAL_NS as TAL
+from chameleon.namespaces import XML_NS
+from chameleon.namespaces import XMLNS_NS
+from chameleon.program import ElementProgram
+from chameleon.utils import ImportableMarker
+from chameleon.utils import decode_htmlentities
 
 
 missing = object()
@@ -69,10 +48,10 @@ def validate_attributes(attributes, namespace, whitelist):
         if ns == namespace and name not in whitelist:
             raise CompilationError(
                 "Bad attribute for namespace '%s'" % ns, name
-                )
+            )
 
 
-def convert_data_attributes(ns_attrs, attrs, namespaces):
+def convert_data_attributes(ns_attrs, attrs, namespaces) -> None:
     d = 0
     for i, attr in list(enumerate(attrs)):
         name = attr['name']
@@ -96,7 +75,7 @@ class MacroProgram(ElementProgram):
         'metal': METAL,
         'i18n': I18N,
         'meta': META,
-        }
+    }
 
     DROP_NS = TAL, METAL, I18N, META
 
@@ -123,13 +102,13 @@ class MacroProgram(ElementProgram):
     # Attributes which should have boolean behavior (on true, the
     # value takes the attribute name, on false, the attribute is
     # dropped)
-    boolean_attributes = set()
+    boolean_attributes: set[str] = set()
 
     # If provided, this should be a set of attributes for implicit
     # translation. Any attribute whose name is included in the set
     # will be translated even without explicit markup. Note that all
     # values should be lowercase strings.
-    implicit_i18n_attributes = set()
+    implicit_i18n_attributes: set[str] = set()
 
     # If set, text will be translated even without explicit markup.
     implicit_i18n_translate = False
@@ -175,7 +154,7 @@ class MacroProgram(ElementProgram):
             'restricted_namespace',
         )
 
-        super(MacroProgram, self).__init__(*args, **kwargs)
+        super().__init__(*args, **kwargs)
 
     @property
     def macros(self):
@@ -185,7 +164,7 @@ class MacroProgram(ElementProgram):
         return tuple(
             nodes.Macro(name, [nodes.Context(node)])
             for name, node in macros
-            )
+        )
 
     def visit_default(self, node):
         return nodes.Text(node)
@@ -241,20 +220,21 @@ class MacroProgram(ElementProgram):
             if use_macro:
                 inner = nodes.UseExternalMacro(
                     nodes.Value(use_macro), slots, False
-                    )
+                )
                 macro_name = use_macro
             else:
                 inner = nodes.UseExternalMacro(
                     nodes.Value(extend_macro), slots, True
-                    )
+                )
                 macro_name = extend_macro
 
-            # While the macro executes, it should have access to the name it was
-            # called with as 'macroname'. Splitting on / mirrors zope.tal and is a
-            # concession to the path expression syntax.
+            # While the macro executes, it should have access to the name it
+            # was called with as 'macroname'. Splitting on / mirrors zope.tal
+            # and is a concession to the path expression syntax.
             macro_name = macro_name.rsplit('/', 1)[-1]
             inner = nodes.Define(
-                [nodes.Assignment(["macroname"], Static(ast.Str(macro_name)), True)],
+                [nodes.Assignment(
+                    ["macroname"], Static(ast.Str(macro_name)), True)],
                 inner,
             )
             STATIC_ATTRIBUTES = None
@@ -276,7 +256,7 @@ class MacroProgram(ElementProgram):
 
                 if end is None:
                     # Make sure start-tag has opening suffix.
-                    start['suffix']  = ">"
+                    start['suffix'] = ">"
 
                     # Explicitly set end-tag.
                     end = {
@@ -284,7 +264,7 @@ class MacroProgram(ElementProgram):
                         'name': start['name'],
                         'space': '',
                         'suffix': '>'
-                        }
+                    }
 
             # i18n:translate
             try:
@@ -317,28 +297,36 @@ class MacroProgram(ElementProgram):
             prepared = tal.prepare_attributes(
                 attrs, TAL_ATTRIBUTES,
                 I18N_ATTRIBUTES, ns, self.DROP_NS
-                )
+            )
 
             # Create attribute nodes
             STATIC_ATTRIBUTES = self._create_static_attributes(prepared)
-            ATTRIBUTES = self._create_attributes_nodes(
-                prepared, I18N_ATTRIBUTES, STATIC_ATTRIBUTES,
+            attributes, filtering = self._create_attributes_nodes(
+                prepared, I18N_ATTRIBUTES
             )
 
+            ATTRIBUTES = nodes.Sequence(attributes)
+
+            # We're caching all expressions found during attribute processing
+            # to enable the filtering functionality which exists to allow later
+            # definitions to override previous ones.
+            if filtering:
+                ATTRIBUTES = nodes.Cache(filtering, ATTRIBUTES)
+
             # Start- and end nodes
             start_tag = nodes.Start(
                 start['name'],
                 self._maybe_trim(start['prefix']),
                 self._maybe_trim(start['suffix']),
                 ATTRIBUTES
-                )
+            )
 
             end_tag = nodes.End(
                 end['name'],
                 end['space'],
                 self._maybe_trim(end['prefix']),
                 self._maybe_trim(end['suffix']),
-                ) if end is not None else None
+            ) if end is not None else None
 
             # tal:omit-tag
             try:
@@ -367,7 +355,7 @@ class MacroProgram(ElementProgram):
                     start_tag,
                     end_tag,
                     content,
-                    )
+                )
 
                 if omit is not False:
                     inner = nodes.Cache([omit], inner)
@@ -403,22 +391,6 @@ class MacroProgram(ElementProgram):
             if defines is None:
                 raise ParseError("Invalid define syntax.", clause)
 
-        # i18n:target
-        try:
-            target_language = ns[I18N, 'target']
-        except KeyError:
-            pass
-        else:
-            # The name "default" is an alias for the target language
-            # variable. We simply replace it.
-            target_language = target_language.replace(
-                'default', 'target_language'
-            )
-
-            defines.append(
-                ('local', ("target_language", ), target_language)
-            )
-
         assignments = [
             nodes.Assignment(
                 names, nodes.Value(expr), context == "local")
@@ -426,7 +398,8 @@ class MacroProgram(ElementProgram):
         ]
 
         # Assign static attributes dictionary to "attrs" value
-        assignments.insert(0, nodes.Alias(["attrs"], STATIC_ATTRIBUTES or EMPTY_DICT))
+        assignments.insert(
+            0, nodes.Alias(["attrs"], STATIC_ATTRIBUTES or EMPTY_DICT))
         DEFINE = partial(nodes.Define, assignments)
 
         # tal:case
@@ -442,20 +415,23 @@ class MacroProgram(ElementProgram):
             else:
                 raise LanguageError(
                     "Must define switch on a parent element.", clause
-                    )
+                )
 
-            CASE = lambda node: nodes.Define(
-                [nodes.Alias(["default"], self.default_marker)],
-                nodes.Condition(
-                    nodes.And([
-                        nodes.BinOp(switch, nodes.IsNot, self._cancel_marker),
-                        nodes.Or([
-                            nodes.BinOp(value, nodes.Equals, switch),
-                            nodes.BinOp(value, nodes.Equals, self.default_marker)
-                        ])
-                    ]),
-                    nodes.Cancel([switch], node, self._cancel_marker),
-                ))
+            def CASE(node):
+                return nodes.Define(
+                    [nodes.Alias(["default"], self.default_marker)],
+                    nodes.Condition(
+                        nodes.And([
+                            nodes.BinOp(
+                                switch, nodes.IsNot, self._cancel_marker),
+                            nodes.Or([
+                                nodes.BinOp(value, nodes.Equals, switch),
+                                nodes.BinOp(
+                                    value, nodes.Equals, self.default_marker)
+                            ])
+                        ]),
+                        nodes.Cancel([switch], node, self._cancel_marker),
+                    ))
 
         # tal:repeat
         try:
@@ -480,7 +456,7 @@ class MacroProgram(ElementProgram):
                 expression,
                 context == "local",
                 whitespace
-                )
+            )
 
         # tal:condition
         try:
@@ -513,6 +489,17 @@ class MacroProgram(ElementProgram):
         else:
             CONTEXT = partial(nodes.TxContext, clause)
 
+        # i18n:target
+        try:
+            clause = ns[I18N, 'target']
+        except KeyError:
+            TARGET = skip
+        else:
+            TARGET = lambda node: nodes.Define(  # noqa:  E731 do not assign a lambda expression, use a def
+                [nodes.Alias(["default"], "target_language")],
+                nodes.Target(clause, node)
+            )
+
         # i18n:name
         try:
             clause = ns[I18N, 'name']
@@ -536,7 +523,8 @@ class MacroProgram(ElementProgram):
             SWITCH,
             DOMAIN,
             CONTEXT,
-            )
+            TARGET,
+        )
 
         # metal:fill-slot
         try:
@@ -558,7 +546,7 @@ class MacroProgram(ElementProgram):
                 raise LanguageError(
                     "Cannot use metal:fill-slot without metal:use-macro.",
                     clause
-                    )
+                )
 
             slots = self._use_macro[index]
             slots.append(nodes.FillSlot(clause, slot))
@@ -576,13 +564,13 @@ class MacroProgram(ElementProgram):
                     clause
                 )
 
-            self._macros[clause] =  slot
+            self._macros[clause] = slot
             slot = nodes.UseInternalMacro(clause)
 
         slot = wrap(
             slot,
             NAME
-            )
+        )
 
         # tal:on-error
         try:
@@ -597,20 +585,20 @@ class MacroProgram(ElementProgram):
             )
 
             if omit is False and start['namespace'] not in self.DROP_NS:
-                start_tag = copy(start_tag)
-
-                start_tag.attributes = nodes.Sequence(
-                    start_tag.attributes.extract(
-                        lambda attribute:
-                        isinstance(attribute, nodes.Attribute) and
-                        isinstance(attribute.expression, ast.Str)
+                start_tag = nodes.Start(
+                    start['name'],
+                    self._maybe_trim(start['prefix']),
+                    self._maybe_trim(start['suffix']),
+                    nodes.Sequence(
+                        [attr for attr in attributes if
+                         isinstance(attr, nodes.Attribute) and
+                         isinstance(attr.expression, ast.Str)]
                     )
                 )
-
                 if end_tag is None:
                     # Make sure start-tag has opening suffix. We don't
                     # allow self-closing element here.
-                    start_tag.suffix  = ">"
+                    start_tag.suffix = ">"
 
                     # Explicitly set end-tag.
                     end_tag = nodes.End(start_tag.name, '', '</', '>',)
@@ -648,13 +636,13 @@ class MacroProgram(ElementProgram):
         return wrap(
             slot,
             ON_ERROR
-            )
+        )
 
     def visit_start_tag(self, start):
         return self.visit_element(start, None, [])
 
     def visit_cdata(self, node):
-        if not self._interpolation[-1] or not '${' in node:
+        if not self._interpolation[-1] or '${' not in node:
             return nodes.Text(node)
 
         expr = nodes.Substitution(node, ())
@@ -670,7 +658,7 @@ class MacroProgram(ElementProgram):
         if node.startswith('<!--?'):
             return nodes.Text('<!--' + node.lstrip('<!-?'))
 
-        if not self._interpolation[-1] or not '${' in node:
+        if not self._interpolation[-1] or '${' not in node:
             return nodes.Text(node)
 
         char_escape = ('&', '<', '>') if self.escape else ()
@@ -728,7 +716,7 @@ class MacroProgram(ElementProgram):
                 "Dynamic attributes not allowed on elements of "
                 "the namespace: %s." % namespace,
                 ns[TAL, 'attributes'],
-                )
+            )
 
         script = ns.get((TAL, 'script'))
         if script is not None:
@@ -740,13 +728,13 @@ class MacroProgram(ElementProgram):
             raise LanguageError(
                 "You cannot use tal:content and tal:replace at the same time.",
                 tal_content
-                )
+            )
 
         if tal_content and ns.get((I18N, 'translate')):
             raise LanguageError(
                 "You cannot use tal:content with non-trivial i18n:translate.",
                 tal_content
-                )
+            )
 
     def _make_content_node(self, expression, default, key, translate):
         value = nodes.Value(expression)
@@ -758,7 +746,7 @@ class MacroProgram(ElementProgram):
                 nodes.BinOp(value, nodes.Is, self.default_marker),
                 default,
                 content,
-                )
+            )
 
             # Cache expression to avoid duplicate evaluation
             content = nodes.Cache([value], content)
@@ -767,11 +755,11 @@ class MacroProgram(ElementProgram):
             content = nodes.Define(
                 [nodes.Alias(["default"], self.default_marker)],
                 content
-                )
+            )
 
         return content
 
-    def _create_attributes_nodes(self, prepared, I18N_ATTRIBUTES, STATIC):
+    def _create_attributes_nodes(self, prepared, I18N_ATTRIBUTES):
         attributes = []
 
         names = [attr[0] for attr in prepared]
@@ -789,11 +777,24 @@ class MacroProgram(ElementProgram):
             # If (by heuristic) ``text`` contains one or more
             # interpolation expressions, apply interpolation
             # substitution to the text.
+            boolean = name in self.boolean_attributes
             if expr is None and text is not None and '${' in text:
                 default = None
-                expr = nodes.Substitution(text, char_escape, default, self.default_marker)
-                translation = implicit_i18n and msgid is missing
+                expr = nodes.Substitution(
+                    text,
+                    char_escape,
+                    default,
+                    self.default_marker,
+                    literal_false=not boolean
+                )
+                translation = (
+                    implicit_i18n and
+                    msgid is missing and
+                    not boolean
+                )
                 value = nodes.Interpolation(expr, True, translation)
+                if boolean:
+                    value = nodes.Replace(value, name)
             else:
                 default = ast.Str(s=text) if text is not None else None
 
@@ -807,20 +808,25 @@ class MacroProgram(ElementProgram):
                             self.default_marker
                         )
                         value = nodes.DictAttributes(
-                            expression, ('&', '<', '>', '"'), '"',
-                            set(filter(None, names[i:]))
+                            expression,
+                            ('&', '<', '>', '"'),
+                            '"',
+                            set(filter(None, names[i:])),
+                            self.boolean_attributes
                         )
                         for fs in filtering:
                             fs.append(expression)
                         filtering.append([])
                     elif name in self.boolean_attributes:
-                        value = nodes.Boolean(expr, name, default, self.default_marker)
+                        value = nodes.Boolean(
+                            expr, name, default, self.default_marker)
                     else:
                         value = nodes.Substitution(
                             decode_htmlentities(expr),
                             char_escape,
                             default,
                             self.default_marker,
+                            literal_false=not boolean
                         )
 
                 # Otherwise, it's a static attribute. We don't include it
@@ -861,16 +867,7 @@ class MacroProgram(ElementProgram):
 
             attributes.append(value)
 
-        result = nodes.Sequence(attributes)
-
-        # We're caching all expressions found during attribute processing to
-        # enable the filtering functionality which exists to allow later
-        # definitions to override previous ones.
-        expressions = filtering[0]
-        if expressions:
-            return nodes.Cache(expressions, result)
-
-        return result
+        return (attributes, filtering[0])
 
     def _create_static_attributes(self, prepared):
         static_attrs = {}
